/*
 * auditd-plugin-clickhouse-lite is a lightweight plugin for auditd
 * daemon to send audit data to a Clickhouse database.
 *
 * Copyright (C) 2023  BaseALT /basealt.ru/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include <signal.h>
#include <poll.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include <algorithm>
#include <string>
#include <map>
#include <fstream>
#include <sstream>
#include <vector>
#include <set>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <chrono>
#include <utility>
#include <memory>
#include <iostream>
#include <queue>
#include <regex>

#include "client.hpp"

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/optional.hpp>

#include <libaudit.h>

#include "logging.hpp"
#include "utils.hpp"

class Lockable {
public:
    Lockable() = default;
    ~Lockable() = default;

    std::unique_lock<std::mutex> lock();
    std::unique_lock<std::mutex> wait();

    template<class Predicate>
    std::unique_lock<std::mutex> wait(Predicate stop_waiting);

    template<class Rep, class Period>
    std::unique_lock<std::mutex> wait_for(
        const std::chrono::duration<Rep, Period> timeout
    );

    template<class Rep, class Period, class Predicate>
    std::unique_lock<std::mutex> wait_for(
        const std::chrono::duration<Rep, Period>& timeout,
        Predicate stop_waiting
    );

    void notify_one() noexcept;
    void notify_all() noexcept;

protected:
    std::mutex mutex;
    std::condition_variable signal;
};

std::unique_lock<std::mutex> Lockable::lock() {
    std::unique_lock<std::mutex> guard(mutex);
    return guard;
}

void Lockable::notify_one() noexcept {
    signal.notify_one();
}

void Lockable::notify_all() noexcept {
    signal.notify_all();
}

std::unique_lock<std::mutex> Lockable::wait() {
    std::unique_lock<std::mutex> guard(mutex);
    signal.wait(guard);
    return guard;
}

template<class Predicate>
std::unique_lock<std::mutex> Lockable::wait(Predicate stop_waiting) {
    std::unique_lock<std::mutex> guard(mutex);
    signal.wait(guard, stop_waiting);
    return guard;
}

template<class Rep, class Period>
std::unique_lock<std::mutex> Lockable::wait_for(
    const std::chrono::duration<Rep, Period> timeout
) {
    std::unique_lock<std::mutex> guard(mutex);
    signal.wait_for(guard, timeout);
    return guard;
}

template<class Rep, class Period, class Predicate>
std::unique_lock<std::mutex> Lockable::wait_for(
    const std::chrono::duration<Rep, Period>& timeout,
    Predicate stop_waiting
) {
    std::unique_lock<std::mutex> guard(mutex);
    signal.wait_for(guard, timeout, stop_waiting);
    return guard;
}

template<class T>
class LockableQueue: public Lockable {
public:
    LockableQueue(size_t max_size): max_size(max_size) {};
    ~LockableQueue() = default;

    void push(std::shared_ptr<T> elem);
    std::shared_ptr<T> pop();
    size_t size();

protected:
    size_t max_size;
    std::queue<std::shared_ptr<T>> elems;
};

template<class T>
void LockableQueue<T>::push(std::shared_ptr<T> elem) {
    {
        auto guard = wait([&]{ return max_size > elems.size(); });
        elems.push(elem);
    }
}

template<class T>
std::shared_ptr<T> LockableQueue<T>::pop() {
    std::shared_ptr<T> elem = elems.front();
    elems.pop();
    return elem;
}

template<class T>
size_t LockableQueue<T>::size() {
    return elems.size();
}

class BlockQueue: public LockableQueue<clickhouse::Block> {
public:
    BlockQueue(size_t max_size):
        LockableQueue<clickhouse::Block>(max_size) {};
    ~BlockQueue() = default;
};

class Thread {
public:
    void run(bool throw_errs);
    void stop(bool now);
    void join();

protected:
    Thread(const std::string klass, int num);
    virtual ~Thread();

    virtual void thread_function(bool throw_errs) {};
    virtual void notify() {};
    bool terminating();
    bool end_of_work();

    void log(const char *format, ...) __attribute__ ((format (printf, 2, 3)));

private:
    const std::string klass;
    int num;
    std::thread *thread = nullptr;
    std::mutex thread_value_mutex;
    std::mutex thread_state_mutex;
    bool terminate_thread = false;
    bool thread_eow = false;
};

Thread::Thread(const std::string klass, int num):
    klass(klass), num(num) {
}

void Thread::log(const char *format, ...) {
    va_list args;
    va_start(args, format);

    char *msg = NULL;
    int result = vasprintf(&msg, format, args);
    if (result >= 0 && msg != NULL) {
        Logger::write("%s[%i]: %s", klass.c_str(), num, msg);
        free(msg);
    }
}

void Thread::run(bool throw_errs) {
    std::lock_guard<std::mutex> thread_lock(thread_value_mutex);
    if (not thread)
        thread = new std::thread(
            [throw_errs](Thread *self) {
                self->log("Started");
                while (not self->terminating()) {
                    try {
                        self->thread_function(throw_errs);
                        if (self->end_of_work())
                            break;
                    } catch (const std::exception &e) {
                        Logger::write("Error while running thread %s[%i]: %s",
                                      self->klass.c_str(), self->num, e.what());
                        if (throw_errs)
                            throw e;
                        Logger::write("Restarting %s[%i]",
                                      self->klass.c_str(), self->num);
                    }
                }
                self->log("Finished");
            },
            this
        );
}

Thread::~Thread() {
    stop(true);
    join();
}

bool Thread::terminating() {
    std::lock_guard<std::mutex> thread_lock(thread_state_mutex);
    return terminate_thread;
}

bool Thread::end_of_work() {
    std::lock_guard<std::mutex> thread_lock(thread_state_mutex);
    return thread_eow;
}

void Thread::stop(bool now) {
    {
        std::lock_guard<std::mutex> thread_lock(thread_state_mutex);
        if (now)
            terminate_thread = true;
        else
            thread_eow = true;
    }

    notify();
}

void Thread::join() {
    std::lock_guard<std::mutex> thread_lock(thread_value_mutex);
    if (thread) {
        if (thread->joinable())
            thread->join();
        delete thread;
        thread = nullptr;
    }
}

class WriterThread: public Thread {
public:
    class Options {
    public:
        Options() = default;
        ~Options() = default;

        std::string table_name;
        size_t min_block_size;
        unsigned int min_ok_times;
        unsigned int max_db_errors;
        unsigned long db_error_usec;
        unsigned long sys_error_msec;
    };

    WriterThread(int num, BlockQueue &queue,
                 Options options,
                 clickhouse::ClientOptions client_options);
    ~WriterThread() = default;

    Client client;
    Options options;
    BlockQueue &queue;

protected:
    void thread_function(bool throw_errs) override;
    void notify() override;
    unsigned long insert_block(
        std::shared_ptr<clickhouse::Block> block,
        size_t max_block_size,
        bool throw_errs
    );
};

WriterThread::WriterThread(int num, BlockQueue &queue,
                           Options options,
                           clickhouse::ClientOptions client_options):
    Thread("Writer", num), client(client_options), options(options),
    queue(queue)
{
    if (options.table_name.empty())
        throw std::runtime_error("Table name should not be empty!");
    
    if (0 == options.min_block_size)
        options.min_block_size = 1;
}

void WriterThread::notify() {
    queue.notify_all();
}

unsigned long WriterThread::insert_block(
    std::shared_ptr<clickhouse::Block> block,
    size_t max_block_size, bool throw_errs
) {
    if (block->GetRowCount() <= max_block_size) {
        // Insert as is.
        unsigned int db_errors = 0;
        while (db_errors < options.max_db_errors ||
               max_block_size == options.min_block_size) {
            try {
                log("Trying to insert block of %zu records...",
                    block->GetRowCount());
                client.Insert(options.table_name, *block);
                log("Successfully inserted %zu records",
                    block->GetRowCount());
                // Return the unchanged block size limit.
                return max_block_size;
            } catch (clickhouse::ServerException &e) {
                log("Caught ServerException with code %d while inserting the records: %s", e.GetCode(), e.what());
                if (terminating())
                    break;
                if (e.GetCode() == MEMORY_LIMIT_EXCEEDED ||
                    e.GetCode() == CANNOT_ALLOCATE_MEMORY ||
                    e.GetCode() == LIMIT_EXCEEDED)
                {
                    ++db_errors;
                } else {
                    if (throw_errs)
                        throw (std::move(e));
                }
                std::this_thread::sleep_for(
                    std::chrono::microseconds(options.db_error_usec)
                );
            } catch (const std::system_error &e) {
                log("Caught %s system error with code %d (%d) while inserting the records: %s", e.code().default_error_condition().category().name(), e.code().value(), e.code().default_error_condition().value(), e.what());
                if (throw_errs)
                    throw e;
                if (terminating())
                    break;
                std::this_thread::sleep_for(
                    std::chrono::milliseconds(options.sys_error_msec)
                );
            } catch (const std::exception &e) {
                log("Caught exception while inserting the records: %s",
                    e.what());
                if (throw_errs)
                    throw e;
                if (terminating())
                    break;
                std::this_thread::sleep_for(
                    std::chrono::milliseconds(options.sys_error_msec)
                );
            }
        }

        // Return due to thread stop().
        if (terminating())
            return max_block_size;

        max_block_size = block->GetRowCount() / 2;
        if (max_block_size < options.min_block_size)
            max_block_size = options.min_block_size;

        log("Reduce block size limit to %zu", max_block_size);

        // Return the reduced block size limit.
        return insert_block(block, max_block_size, throw_errs);
    }

    // Slicing the block on max_block_size blocks.
    size_t pos = 0;
    size_t block_size_sum = 0;
    unsigned int inserts = 0;

    log("Slicing the block into blocks of %zu records...",
        max_block_size);
    
    while (pos < block->GetRowCount()) {
        size_t count = max_block_size;
        if (pos + count > block->GetRowCount())
            count = block->GetRowCount() - pos;

        std::shared_ptr<clickhouse::Block> next_block =
            std::make_shared<clickhouse::Block>(
                block->GetColumnCount(), count
            );

        for (auto iter = block->begin(); iter != block->end(); ++iter) {
            next_block->AppendColumn(iter.Name(),
                                     iter.Column()->Slice(pos, count));
        }

        block_size_sum += insert_block(next_block, max_block_size, throw_errs);

        if (terminating())
            break;

        pos += count;
        inserts += 1;
    }

    // Return the average block size limit.
    return block_size_sum / inserts;
}

void WriterThread::thread_function(bool throw_errs) {
    size_t max_block_size = SIZE_MAX;

    unsigned int errors = 0;
    unsigned int ok_times = 0;

    if (throw_errs)
        log("Will throw permament DB and system exceptions");

    log("Block size: unlimited");

    while (not terminating()) {
        std::shared_ptr<clickhouse::Block> block;
        {
            std::unique_lock<std::mutex> guard =
                queue.wait([&]{
                    return queue.size() > 0 || end_of_work();
                });

            if (terminating()) {
                // Wait interrupted by stop(true).
                break;
            }

            if (queue.size() > 0) {
                block = queue.pop();
                queue.notify_all();
            } else if (end_of_work()) {
                // Wait interrupted by stop(false).
                break;
            } else
                continue;
        }

        size_t new_block_size = insert_block(block, max_block_size, throw_errs);

        if (terminating()) {
            // Terminated while inserting.
            break;
        }

        log("Done inserting block of %zu records",
            block->GetRowCount());

        if (new_block_size < max_block_size) {
            max_block_size = new_block_size;
            errors += 1;
            log("Block size: %zu, errors: %u",
                max_block_size, errors);
        } else {
            if (errors > 0)
                errors -= 1;
            if (max_block_size < block->GetRowCount()) {
                ok_times += 1;
                log("Block size: %zu, errors: %u, ok times: %u/%u",
                    max_block_size, errors,
                    ok_times, options.min_ok_times);
                if (ok_times == options.min_ok_times) {
                    max_block_size = max_block_size * 2;
                    ok_times = 0;
                    log("Expand block size limit to %zu",
                        max_block_size);
                }
            }
        }
    }
}

class InputBuffer: public Lockable {
public:
    InputBuffer() = default;
    ~InputBuffer() = default;

    char data[MAX_AUDIT_MESSAGE_LENGTH];
    size_t length = 0;
    void reset();
};

void InputBuffer::reset() {
    length = 0;
}

class SplitterThread: public Thread {
public:
    SplitterThread(int num, InputBuffer &buf, BlockQueue &queue,
                   size_t block_size, unsigned int timeout_sec);

protected:
    void thread_function(bool throw_errs) override;
    void notify() override;
    InputBuffer &buf;
    BlockQueue &queue;
    size_t block_size;
    unsigned int timeout_sec;
};

SplitterThread::SplitterThread(int num, InputBuffer &buf,
                               BlockQueue &queue,
                               size_t block_size,
                               unsigned int timeout_sec):
    Thread("Splitter", num), buf(buf), queue(queue),
    block_size(block_size), timeout_sec(timeout_sec) {
}

void SplitterThread::notify() {
    buf.notify_all();
}

void SplitterThread::thread_function(bool throw_errs) {
    auto record_text_col = std::make_shared<clickhouse::ColumnString>();

    auto push_block = [&]() {
        auto block = std::make_shared<clickhouse::Block>();
        block->AppendColumn("record_text", record_text_col);
        queue.push(block);
        queue.notify_one();
        record_text_col =
            std::make_shared<clickhouse::ColumnString>();
    };

    log("Using timeout of %u s", timeout_sec);
    auto timeout = std::chrono::seconds(timeout_sec);

    while (not terminating()) {
        std::unique_lock<std::mutex> guard;

        if (record_text_col->Size() == 0 || timeout_sec == 0)
            guard = buf.wait([&]{
                return buf.length > 0 || end_of_work();
            });
        else
            guard = buf.wait_for(timeout, [&]{
                return buf.length > 0 || end_of_work();
            });

        if (terminating()) {
            // Wait interrupted by stop(true).
            break;
        }

        if (buf.length > 0) {
            // New data.
            buf.data[buf.length - 1] = '\0';
            char *rest = buf.data;
            do {
                const char *record = strsep(&rest, "\n");
                record_text_col->Append(record);
                if (record_text_col->Size() == block_size) {
                    // Size limit.
                    push_block();
                }
            } while (rest);
            buf.reset();
            guard.unlock();
            if (not end_of_work()) {
                buf.notify_one();
            } else if (record_text_col->Size() > 0) {
                push_block();
                break;
            }
        } else {
            // No data, timeout or end of work.
            guard.unlock();
            if (record_text_col->Size() > 0) {
                push_block();
            }
            if (end_of_work()) {
                // Wait interrupted by stop(false).
                break;
            }
        }
    }
}

bool got_term = false;
static void term_handler(int sig) {
    got_term = true;
}

static int main_func(int argc, char **argv);

int main(int argc, char **argv) {
    try {
        return main_func(argc, argv);
    } catch (const std::exception &e) {
        Logger::write("Caught exception: %s", e.what());
        Logger::set_default_instance(std::make_shared<FileLogger>(""));
        throw e;
    }
}

typedef void* (memcpy_func_t) (void*, const void*, size_t);

static int main_func(int argc, char **argv) {
    if (argc < 2) {
        std::cerr << "Error: Usage: auditd-plugin-clickhouse CONFIG [-e]\n";
        return -1;
    }

    clickhouse::ClientOptions client_options;

    WriterThread::Options writer_options;
    writer_options.table_name = "AuditDataRaw";
    writer_options.min_block_size = 100;
    writer_options.db_error_usec = 10000;
    writer_options.sys_error_msec = 500;
    writer_options.max_db_errors = 5;
    writer_options.min_ok_times = 5;

    std::string init_sql_filename;
    unsigned int timeout_sec = 5;
    size_t max_block_size = 100000;
    unsigned int read_report_sec = 0;

    unsigned int input_buf_count = 8;
    unsigned int num_of_writers = 16;
    size_t queue_depth = 1000;

    bool throw_errs = (argc > 2 && std::string(argv[2]) == "-e");

    {
        struct sigaction sa;

        sa.sa_flags = 0;
        sigemptyset(&sa.sa_mask);

        sa.sa_handler = term_handler;
        if (sigaction(SIGTERM, &sa, NULL) < 0) {
            throw std::runtime_error("Failed to set SIGTERM handler");
        }
    }

    {
        boost::property_tree::ptree config_tree;
        std::ifstream config(argv[1]);
        boost::property_tree::read_ini(config, config_tree);

        auto general_sec = config_tree.get_child_optional("General");
        if (general_sec) {
            optional_set(timeout_sec, *general_sec,
                         "Timeout");
            optional_set(max_block_size, *general_sec,
                         "MaxBlockSize");
            optional_set(input_buf_count, *general_sec,
                         "InputBufCount");
            optional_set(num_of_writers, *general_sec,
                         "NumOfWriters");
            optional_set(queue_depth, *general_sec,
                         "QueueDepth");

            if (writer_options.min_block_size == 0)
                writer_options.min_block_size = 1;

            if (max_block_size < writer_options.min_block_size)
                max_block_size = writer_options.min_block_size;
        }

        auto reliability_sec =
            config_tree.get_child_optional("Reliability");
        if (reliability_sec) {
            optional_set(writer_options.min_block_size, *reliability_sec,
                         "MinBlockSize");
            optional_set(writer_options.max_db_errors, *reliability_sec,
                         "MaxDatabaseErrors");
            optional_set(writer_options.db_error_usec, *reliability_sec,
                         "DatabaseErrorDelay");
            optional_set(writer_options.db_error_usec, *reliability_sec,
                         "SystemErrorDelay");
            optional_set(writer_options.min_ok_times, *reliability_sec,
                         "MinOKTimes");
        }

        auto database_sec =
            config_tree.get_child_optional("Database");
        if (database_sec) {
            optional_set(init_sql_filename, *database_sec,
                         "InitSQLFile");
            optional_set(writer_options.table_name, *database_sec,
                         "TableName");
            optional_set(client_options.host, *database_sec,
                         "Hostname");
            optional_set(client_options.port, *database_sec,
                         "Port");
            optional_set(client_options.default_database,
                         *database_sec,
                         "Database");
            optional_set(client_options.user, *database_sec,
                         "Username");
            optional_set(client_options.password, *database_sec,
                        "Password");
            optional_set(client_options.ping_before_query,
                         *database_sec,
                         "PingBeforeQuery");
            optional_set(client_options.send_retries, *database_sec,
                         "SendRetries");

            {
                auto retry_timeout =
                    database_sec->get_child_optional("RetryTimeout");
                if (retry_timeout) {
                    client_options.retry_timeout =
                        std::chrono::seconds(retry_timeout->get_value<int>());
                }
            }

            {
                std::string compression_method;
                optional_set(compression_method, *database_sec,
                             "CompressionMethod");
                if (compression_method == "None")
                    client_options.compression_method =
                        clickhouse::CompressionMethod::None;
                else if (compression_method == "LZ4")
                    client_options.compression_method =
                        clickhouse::CompressionMethod::LZ4;
                else
                    throw std::runtime_error(
                        "Unexpected compression method"
                    );
            }
        }

        auto logging_sec =
            config_tree.get_child_optional("Logging");
        if (logging_sec) {
            std::string method;
            optional_set(method, *logging_sec, "Method");

            std::string logfilename;
            optional_set(logfilename, *logging_sec, "Logfile");

            std::string syslogname;
            optional_set(syslogname, *logging_sec, "Ident");

            if (method.empty() || method == "stream") {
                Logger::set_instance(std::make_shared<FileLogger>(logfilename.c_str()));
            } else if (method == "stderr") {
                Logger::set_instance(std::make_shared<FileLogger>(""));
            } else if (method == "syslog") {
                Logger::set_instance(std::make_shared<SyslogLogger>(syslogname.empty() ? NULL : syslogname.c_str()));
            } else {
                throw std::runtime_error("Unexpected logging method");
            }

            optional_set(read_report_sec, *logging_sec,
                         "ReadReportPeriod");
        }
    }

    Logger::set_default_instance(std::make_shared<FileLogger>(""));

    Logger::write("Initializing");

    if (not init_sql_filename.empty()) {
        Logger::write("Run initial SQL script %s",
                      init_sql_filename.c_str());
        Client client(client_options);
        std::ifstream init_sql_file(init_sql_filename);
        std::string init_sql(
            std::istreambuf_iterator<char>{init_sql_file},
            {}
        );
        std::regex stmt_end(";\\n+");
        std::sregex_token_iterator iter(
            init_sql.begin(), init_sql.end(), stmt_end, -1
        );
        std::sregex_token_iterator end;
        for (; iter != end; ++iter) {
            client.Execute(std::string(*iter));
        }
    }

    Logger::write("Setting up input buffers, block queue and threads...");

    BlockQueue queue(queue_depth);
    std::vector<InputBuffer> in_buffers(input_buf_count);
    std::vector<std::shared_ptr<SplitterThread>> splitters;

    for (size_t i = 0; i < in_buffers.size(); ++i) {
        auto s = std::make_shared<SplitterThread>(
            i + 1, in_buffers[i], queue,
            max_block_size, timeout_sec
        );
        s->run(throw_errs);
        splitters.push_back(std::move(s));
    }

    std::vector<std::shared_ptr<WriterThread>> writers;
    for (unsigned int i = 0; i < num_of_writers; ++i) {
        auto w = std::make_shared<WriterThread>(
            i + 1, queue, writer_options, client_options
        );
        w->run(throw_errs);
        writers.push_back(std::move(w));
    }

    Logger::write("Initialization finished");
    
    struct pollfd pollfds[1];
    pollfds[0].fd = STDIN_FILENO;

    auto read_start_time = std::chrono::steady_clock::now();
    auto last_read_time = read_start_time;
    unsigned long long last_read_size = 0;
    unsigned long long total_read = 0;

	const char *tail = NULL;
    size_t tail_len = 0;

    unsigned int buf_num = 0;
    while (not got_term) {
        pollfds[0].events = POLLIN | POLLHUP;
        pollfds[0].revents = 0;

        int res = poll(
            pollfds, sizeof(pollfds) / sizeof(pollfds[0]),
            -1 // wait forever
        );

        if (res < 0) {
            if (errno != EINTR) {
                // Error occured, finishing the run.
                Logger::write("Poll returned error: %d", errno);
                break;
            }
        } else if (res == 0) {
            // Timeout? poll() was asked to wait forever...
        } else if (pollfds[0].revents & POLLIN) {
            ssize_t readsize;

            // Buffer lock.
            {
                std::unique_lock<std::mutex> guard =
                    in_buffers[buf_num].wait([&]{
                        return in_buffers[buf_num].length == 0;
                    });

                // Copy the tail from previous buffer (the splitter
                // doesn't use the tail). Use memmove if tail is
				// located in the same buffer as the current.
                if (tail) {
                    memcpy_func_t *cpy_fn = NULL;
                    if (input_buf_count == 1)
                        cpy_fn = &memmove;
                    else
                        cpy_fn = &memcpy;

                    (*cpy_fn)(in_buffers[buf_num].data,
                              tail, tail_len);
                    tail = NULL;
                }

                // Read data into buffer until newline is right-found
                // or there is no space (fatal).
                const char *last_n = NULL;
                char *read_to = in_buffers[buf_num].data + tail_len;
                size_t free_space = sizeof(in_buffers[buf_num].data) - tail_len;
                while (not last_n && free_space > 0) {
                    readsize = read(STDIN_FILENO, read_to, free_space);
                    if (readsize > 0) {
                        last_n = (const char *) memrchr(read_to, '\n', readsize);
                        read_to += readsize;
                        free_space -= readsize;
                        last_read_size += readsize;
                    } else {
                        break;
                    }
                }

                // The new tail is bytes after the last newline.
                // It's fatal if some data wthere is still no newline
                // within the available data.
                if (last_n) {
                    in_buffers[buf_num].length =
                        (last_n - in_buffers[buf_num].data) + 1;
                    tail_len = read_to - last_n - 1;
                    if (tail_len)
                        tail = last_n + 1;
                } else {
                    if (free_space != sizeof(in_buffers[buf_num].data)) {
                        throw std::runtime_error(
                            "Critical error: No record end marker found within input buffer!"
                        );
                    }
                }
            }
            // Buffer unlock.

            if (in_buffers[buf_num].length > 0)
                in_buffers[buf_num].notify_one();

            if (read_report_sec > 0) {
                auto current_read_time =
                    std::chrono::steady_clock::now();
                unsigned long read_elapsed =
                    std::chrono::duration_cast<std::chrono::seconds>(
                        current_read_time - last_read_time
                    ).count();
                if (read_elapsed >= read_report_sec) {
                    Logger::write("Read %llu bytes in last %lu seconds",
                                  last_read_size, read_elapsed);
                    total_read += last_read_size;
                    last_read_size = 0;
                    last_read_time = current_read_time;
                }
            }

            if (0 == readsize) {
                Logger::write("End of input. Finishing the run...");
                break;
            } else if (readsize < 0) {
                Logger::write("Read error (%d). Finishing the run...",
                              errno);
                break;
            }
        } else if (pollfds[0].revents & POLLHUP) {
            Logger::write("Hangup. Finishing the run...");
            break;
        } else if (pollfds[0].revents & POLLERR) {
            Logger::write("Input error (%d). Finishing the run...",
                          errno);
            break;
        }

        buf_num += 1;
        if (buf_num == input_buf_count)
            buf_num = 0;
    }

    total_read += last_read_size;

    auto read_end_time = std::chrono::steady_clock::now();
    unsigned long total_elapsed =
        std::chrono::duration_cast<std::chrono::seconds>(
            read_end_time - read_start_time
        ).count();

    Logger::write("Totally read %llu bytes in %lu seconds",
                  total_read, total_elapsed);

    Logger::write("Setting down input buffers, block queue and threads...");

    for (auto iter = splitters.begin(); iter != splitters.end(); ++iter) {
        iter->get()->stop(false);
    }
    for (auto iter = splitters.begin(); iter != splitters.end(); ++iter) {
        iter->get()->join();
    }

    for (auto iter = writers.begin(); iter != writers.end(); ++iter) {
        iter->get()->stop(false);
    }
    for (auto iter = writers.begin(); iter != writers.end(); ++iter) {
        iter->get()->join();
    }

    Logger::write("All threads finished. Exit");

    return 0;
}
