/*
 * auditd-plugin-clickhouse-lite is a lightweight plugin for auditd
 * daemon to send audit data to a Clickhouse database.
 *
 * Copyright (C) 2023  BaseALT /basealt.ru/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "logging.hpp"
#include <time.h>
#include <syslog.h>
#include <string.h>

std::list<std::pair<std::string, std::string> > Logger::s_lines;
std::recursive_mutex Logger::s_mutex;
std::shared_ptr<LoggerImpl> Logger::s_logger;

Logger main_logger;

bool Logger::has_instance()
{
    std::lock_guard<std::recursive_mutex> lock(Logger::s_mutex);
    return Logger::s_logger ? true : false;
}

void Logger::set_instance(std::shared_ptr<LoggerImpl> logger)
{
    std::lock_guard<std::recursive_mutex> lock(Logger::s_mutex);
    Logger::s_logger = logger;
    Logger::dump();
}

void Logger::set_default_instance(std::shared_ptr<LoggerImpl> logger)
{
    std::lock_guard<std::recursive_mutex> lock(Logger::s_mutex);
    if (not Logger::s_logger) {
        Logger::s_logger = logger;
        Logger::dump();
    }
}

void Logger::write(const char *format, ...)
{
    std::lock_guard<std::recursive_mutex> lock(Logger::s_mutex);

    Logger::dump();

    std::string msg;

    va_list args;
    va_start(args, format);
    char *buf = NULL;
    int result = vasprintf(&buf, format, args);
    if (result >= 0 && buf != NULL)
        msg = buf;
    free(buf);
    va_end(args);

    if (msg.empty())
        return;

    if (Logger::s_logger) {
        Logger::s_logger->log(msg);
    } else {
        Logger::s_lines.push_back(std::make_pair(getdatestring(), msg));
    }
}

void Logger::dump()
{
    std::lock_guard<std::recursive_mutex> lock(Logger::s_mutex);

    if (Logger::s_logger && s_lines.size() > 0) {
        for (auto iter = Logger::s_lines.begin();
             iter != Logger::s_lines.end(); ++iter)
        {
	    Logger::s_logger->log(iter->first, iter->second);
        }

        Logger::s_lines.clear();
    }
}

Logger::~Logger() {
    Logger::set_default_instance(std::make_shared<FileLogger>(""));
}

FileLogger::FileLogger(const char *filename)
{
    m_logfile_name = filename;
}

void FileLogger::log(std::string timestamp, std::string message)
{
    if (!m_logfile_name.empty()) {
        FILE *outfile = NULL;
        outfile = fopen(m_logfile_name.c_str(), "at");

        if (outfile != NULL) {
            fprintf(outfile, "%s: %s\n", timestamp.c_str(),
                    message.c_str());
            fclose(outfile);
        } else {
            fprintf(stderr,
                    "LOG ERROR! Unable to open %s for append writing!\n",
                    m_logfile_name.c_str());
        }
    } else {
        fprintf(stderr, "%s: %s\n", timestamp.c_str(),
                message.c_str());
        fflush(stderr);
    }
}

void FileLogger::log(std::string message)
{
    log(Logger::getdatestring(), message);
}

std::string Logger::getdatestring()
{
    std::string result;

    time_t t = time(nullptr);
    struct tm tm_inst;

    auto tm_ptr = localtime_r(&t, &tm_inst);

    if (tm_ptr != nullptr) {
        char buffer[4096];

        auto buffer_res = strftime(buffer,
                                   sizeof(buffer) - 1,
                                   "%c %Z",
                                   tm_ptr);
        if (buffer_res > 0) {
            buffer[buffer_res] = 0;
            result = buffer;
        }
    }

    return result;
}

SyslogLogger::SyslogLogger(const char *ident)
{
    openlog(ident, 0, LOG_DAEMON);
}

SyslogLogger::~SyslogLogger()
{
    closelog();
}

void SyslogLogger::log(std::string timestamp, std::string message)
{
    syslog(LOG_INFO, "%s: %s", timestamp.c_str(), message.c_str());
}

void SyslogLogger::log(std::string message)
{
    syslog(LOG_INFO, "%s", message.c_str());
}
